
(* le type direction *)


let est = 0
and nord = 1
and ouest = 2 
and sud = 3;;

let oppose d = match d with 
    0 -> 2
  | 2  -> 0
  | 1 -> 3
  | 3 -> 1
  | _ -> 0
;;

(* Les classes abstraites *)

class virtual lieu () =
object
  method virtual entrer : unit -> bool
  method virtual to_lieu : lieu
end;;

class virtual salle () = 
object(self)
  inherit lieu ()
  method virtual installeCote : int -> lieu -> unit
  method virtual acqCote : int -> lieu 
  method virtual acqNumero : unit -> int
  method virtual acqDirection  : lieu -> int 
  method to_lieu  = upcast self from salle to lieu
end;;

class virtual mur () = 
object(self)
  inherit lieu ()
  method virtual poserMur : int -> salle -> unit
  method virtual acqSalle : unit -> salle
  method virtual acqDirection : unit -> int 
  method virtual accolleMur : mur -> unit
  method virtual acqMitoyen : unit -> mur
  method to_lieu = upcast self from mur to lieu
end;;


class virtual porte () = 
object(self)
  inherit lieu ()
  method virtual poserPorte : int -> salle -> salle -> bool -> unit
  method virtual acqDirection : salle -> int
  method virtual acqAutreCote : salle -> salle
  method to_lieu = upcast self from porte to lieu
end;;


(* Les classes concretes *)

class lieu_concret () =
object (self)
  inherit lieu () 
  method entrer () = true
  method to_lieu = upcast self from lieu_concret to lieu
end;;

let lc1 = new lieu_concret ();;


class salleNulle num = 
object(self) 
  inherit salle () 
  method to_salle = (upcast self from salleNulle to salle)
(*   method to_lieu = (self#to_salle)#to_lieu *)
  method entrer () = true
  method installeCote (i : int) (l : lieu)  = ()
  method acqNumero () = num
  method acqCote (i : int) = (self#to_lieu)
  method acqDirection (l : lieu) = -1
end;;

let salleNulle = new salleNulle (-1);;
  
exception Found of int;;

class salleNormale num = 
object(self)
  inherit salleNulle num
  val mutable numero = num 
  val mutable cotes = [||]
  initializer cotes <- Array.create 4 (self#to_lieu)
  method to_salleNulle = upcast self from salleNormale to salleNulle
(*  method to_lieu = self#to_salleNulle#to_lieu*)
  method entrer () = true
  method installeCote (i : int) (l : lieu)  = cotes.(i) <- l
  method acqNumero () = numero
  method acqCote (i : int) = cotes.(i)
  method acqDirection (l : lieu) = 
    try 
      for i=0 to 3 do
        if l == cotes.(i) then raise (Found i) 
      done;
      -1
    with Found i -> i  

end;;

class murNul d s = 
object(self)
  inherit lieu () 
  inherit mur () 
  method to_lieu = upcast self from murNul to lieu
  method to_mur = upcast self from murNul to mur
  method poserMur (d : int) (s : salle) =   ()
  method entrer () = false
  method acqSalle () = s
  method acqDirection () = d
  method accolleMur (m : mur) = ()
  method acqMitoyen () = upcast self from murNul to  mur
  method to_lieu = upcast self from murNul to lieu
  method to_mur = upcast self from murNul to mur
end;;

let murNul = new murNul (-1) (upcast salleNulle from salleNulle to  salle);;

class murNormal d s = 
object(self)
  inherit murNul d s
  inherit mur () 
  method to_mur = upcast self from murNormal to mur 
  val mutable direction = d
  val mutable salle = s
  val mutable mitoyen =  (upcast murNul from murNul to  mur)
  initializer s#installeCote d 
     (upcast (upcast self from murNormal to mur) from mur to lieu)

  method poserMur (d : int) (s : salle) = 
   direction <- d;
   salle <- s;
   s#installeCote d 
     (upcast (upcast self from murNormal to mur) from mur to lieu)

  method entrer () = false
  method acqSalle () = salle
  method acqDirection () = direction
  method accolleMur (m : mur) = mitoyen <- m  
  method acqMitoyen () = mitoyen
end;;


class porteNormale d s1 s2 b = 
object(self)
  inherit lieu () 
  inherit porte()
  val mutable direction = d
  val mutable salle1 = s1
  val mutable salle2 = s2
  val mutable ouverte = b
  initializer 
    salle1#installeCote direction (upcast self from porteNormale to  lieu);
    salle2#installeCote (oppose direction) (upcast self from porteNormale to lieu) 
  method poserPorte (d : int) (s1 : salle) (s2 : salle) ( b : bool) = 
   direction <- d; 
   salle1 <- s1;
   salle2 <- s2;
   ouverte <- b;  
   salle1#installeCote direction (upcast self from porteNormale to  lieu);
   salle2#installeCote (oppose direction)  (upcast self from porteNormale to  lieu) 
  method entrer () = ouverte
  method acqDirection(s : salle) = 
    if ( s == salle1) then direction 
    else if (s == salle2) then (oppose direction)
    else -1
  method acqAutreCote (s : salle) = 
    if (s == salle1) then salle2
    else if (s == salle2) then salle1
    else  (upcast salleNulle from salleNulle to  salle)
end;;

exception Break;;

class dedale () =
object(self)
  val mutable plan = ( [] : salle list)
  method ajouteSalle (s : salle) = plan <- s :: plan
  method accolleSalles ( d1 : int) (s1 : salle) (s2 : salle) = 
    let ac1 = s1#acqCote(d1) 
    and ac2 = s2#acqCote(oppose d1) in 
      let m1 = downcast ac1 from lieu to mur
      and m2 = downcast ac2 from lieu to mur in 
        m1#accolleMur(m2);
        m2#accolleMur(m1)
  method  salleNumero (num : int) = 
    let p_aux  = ref plan  
    and res = ref ( upcast salleNulle from salleNulle to  salle) in 
    try 
      while ( !p_aux <> []) do 
        let t = List.hd !p_aux in 
          if t#acqNumero() == num then ( res := t; raise Break)
          else p_aux := List.tl (!p_aux)
      done;
      !res
	  with Break -> !res
end;;


class virtual fabriqueDedale () = 
object
  method virtual creeDedale : unit -> dedale
  method virtual creeMur : int -> salle -> mur
  method virtual creeSalle : int -> salle
  method virtual creePorte : int -> salle -> salle -> bool -> porte
end;;


class fabriqueDedaleNormal () =
object
  inherit fabriqueDedale ()
  method creeDedale () = new dedale ()
  method creeMur d s = (new murNormal d s)#to_mur
  method creeSalle n = (new salleNormale n)#to_salle
  method creePorte d s1 s2 b = new porteNormale d s1 s2 b
end;;


class murBombarde d s = 
object 
  inherit murNormal d s
  val mutable bombarde = false
  method endommager () = bombarde <- true
  method entrer () = bombarde
end


class sallePiege num = 
object(self)
  inherit salleNormale num
  val mutable alea = false
  initializer Random.init 10; if Random.float 1.0 < 0.25 then alea <- true
  method entrer () = 
    if alea then (self#propagerExplosion(); alea <- false; alea)
    else true
method propagerExplosion () = 
  for i=0 to 3 do
    try 
      let mb = downcast cotes.(i) from lieu to murBombarde in 
      mb#endommager()
    with Cast.Cast_failure (s,_,_) -> () 
  done
end;;


class fabriqueDedalePiege () =
object
  method creeDedale () = new dedale ()
  method creeMur d s = 
   upcast ( upcast (new murBombarde d s) from murBombarde to murNormal)
   from murNormal to mur
  method creeSalle n = 
   upcast 
     (upcast ( upcast (new sallePiege n) from sallePiege to salleNormale)
      from salleNormale to  salleNulle)
   from salleNulle to salle
  method creePorte d s1 s2 b = 
    upcast ((new porteNormale d s1 s2 b)) from porteNormale to porte
end;;


class jeuDuLabyrinthe () = 
object
  method creeDedale ( fabrique : fabriqueDedale)  = 
    let unDedale = fabrique#creeDedale()  in 
    for i=1 to 3 do 
      let s = fabrique#creeSalle  i in 
        unDedale#ajouteSalle(s);
        ignore(fabrique#creeMur est s);
        ignore(fabrique#creeMur ouest s);
        ignore(fabrique#creeMur sud s);
        ignore(fabrique#creeMur nord s)
    done;
    let laPorte = 
      fabrique#creePorte est (unDedale#salleNumero 1) 
                         (unDedale#salleNumero 2) true 
    in
      unDedale#accolleSalles nord (unDedale#salleNumero 1) 
                                 (unDedale#salleNumero 3);
      Printf.printf ("dedale cree\n");
     unDedale 
end;;

let main () = 
  Printf.printf "Debut\n";
  let fabriqueN = new fabriqueDedaleNormal () 
  and fabriqueP = new fabriqueDedalePiege () in 
  let jeuN = new jeuDuLabyrinthe ()
  and jeuP = new jeuDuLabyrinthe () in
  ignore(jeuN#creeDedale(fabriqueN));
  ignore(jeuP#creeDedale(fabriqueP));
    Printf.printf "Fin\n";;  

main();;



