(*

1/ original variants (rectangle, circle, translate) and operations (draw) 

              shape
         /      |     \
        /       |      \
       /        |       \
   rectangle  cercle    translate

2/ new variant  (union)

         shape 
        / | \ \
          |
        union

3/ new operation (bounding_box)

        shape 
       // | \\
       /  |
      /   |
 circle  bb_shape
   |   //     \\
   |   /
   |  /
   bb_circle

*)


type window = int;;
   
class virtual shape  =
object
  method virtual draw : window -> int -> int ->  unit
end


class rectangle (w:int) (h:int) = 
object
  inherit shape
  val width = w
  val height = h
  method draw w x y = Printf.printf "affiche sur la fenetre %d un rectangle de taille %d,%d en position %d,%d\n" w width height x y   
end

class circle  (r:int) = 
object
  inherit shape
  val radius = r
  method draw w x y = Printf.printf "affiche sur la fenetre %d un cercle de rayon %d en position %d,%d\n" w radius x y    
end

class translated (s:shape) dx dy = 
object
  inherit shape 
  val delta_x = dx
  val delta_y = dy
  val shape = s
  method draw w x y = s#draw w (x+dx) (y+dy)
end

  let display_shape s = 
     if s instanceof shape then 
        let window = 1 in s#draw window 0 0 
     else 
        failwith "display_shape : not a shape";;


  display_shape (new translated (new rectangle 50 100) 30 30);;

class union (l:'a) (r:'a) = 
object(self:'a) 
  inherit shape
  val left = l 
  val right = r
  method draw w x y = 
    left#draw w x y; 
    right#draw w x y; ()
end;;

  display_shape (new union (new rectangle 10 30) 
                (new translated (new circle 20) 30 30));;
  

type bounding_box = {left:int; bottom:int; right:int;top:int};;

let display_bounding_box bb = 
  Printf.printf "Bounding box %d %d %d %d\n" bb.left bb.bottom bb.right bb.top;;

class virtual bb_shape = 
  object
    inherit shape
    method virtual bounding_box : unit -> bounding_box
  end;;


class bb_rectangle w h = 
object 
  inherit rectangle w h 
  inherit bb_shape 
  method bounding_box () = {left=0;bottom=0;right=width;top= height}
end;;

class bb_circle r = 
object 
  inherit circle r
  inherit bb_shape 
  method bounding_box () = 
    {left= -radius;bottom= -radius;right= radius; top = radius}
end;;

class bb_translated bbs dx dy = 
object 
  inherit translated (bbs:>shape) dx dy
  inherit bb_shape 
  method bounding_box () = let bb = bbs#bounding_box() in 
     {left=bb.left+delta_x; bottom=bb.bottom+delta_y; 
       right=bb.right+delta_x; top=bb.top+delta_y}
end;;

class bb_union (bbs1:'a) (bbs2:'a)  = 
object(self:'a) 
  inherit union bbs1 bbs2
  inherit bb_shape 
  method bounding_box () = 
    let bb1 = bbs1#bounding_box () 
    and bb2 = bbs2#bounding_box () in 
      {left=min bb1.left bb2.left;
       bottom= min bb1.bottom bb2.bottom;
       right = max bb1.right bb2.right;
       top = max bb1.top bb2.top}
end;;


  let display_bb_shape bbs = 
     if bbs instanceof bb_shape then 
        let bb = bbs#bounding_box () in 
          display_bounding_box bb;
          bbs#draw 1 bb.left bb.bottom
     else 
        failwith "display_bb_shape : not a bb shape"
;;

    display_bb_shape (new bb_union (new bb_rectangle 10 30) 
                (new bb_translated (new bb_circle 20) 30 30));;
  




